# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import datetime
from dateutil import parser as date_parser
import fixtures
import mock
from oslo_config import cfg
import sys
import tempfile
import testtools

from subunit2sql import exceptions
from subunit2sql.migrations import cli as migration_cli
from subunit2sql import shell
from subunit2sql.tests import base


class TestShell(base.TestCase):
    def test_run_totals(self):
        fake_results = {}
        # Fake Success
        for num in range(100):
            test_name = 'fake_test_' + str(num)
            fake_results[test_name] = {'status': 'success'}
        # Fake skips
        for num in range(50):
            test_name = 'fake_test_skip_' + str(num)
            fake_results[test_name] = {'status': 'skip'}
        # Fake skips
        for num in range(16):
            test_name = 'fake_test_fail_' + str(num)
            fake_results[test_name] = {'status': 'fail'}
        totals = shell.get_run_totals(fake_results)
        self.assertEqual(totals['success'], 100)
        self.assertEqual(totals['fails'], 16)
        self.assertEqual(totals['skips'], 50)

    def test_running_avg(self):
        fake_test = mock.MagicMock()
        fake_test.success = 150
        fake_test.run_time = 30.452
        fake_result = {
            'start_time': datetime.datetime(1914, 6, 28, 10, 45, 0),
            'end_time': datetime.datetime(1914, 6, 28, 10, 45, 50),
        }
        fake_values = {}
        result = shell.running_avg(fake_test, fake_values, fake_result)
        # Let's do some arithmetic
        expected_avg = ((150 * 30.452) + 50) / 151
        # Both values should be 30.581456953642384
        self.assertEqual(expected_avg, result['run_time'])

    def test_running_avg_no_prev(self):
        fake_test = mock.MagicMock()
        fake_test.success = 1
        fake_test.run_time = None
        fake_result = {
            'start_time': datetime.datetime(1914, 6, 28, 10, 45, 0),
            'end_time': datetime.datetime(1914, 6, 28, 10, 45, 50),
        }
        fake_values = {}
        result = shell.running_avg(fake_test, fake_values, fake_result)
        expected_avg = 50
        self.assertEqual(expected_avg, result['run_time'])

    def test_increment_counts_success(self):
        fake_test = mock.MagicMock()
        fake_test.run_count = 15
        fake_test.success = 5
        fake_test.run_time = 45.0
        fake_result = {
            'status': 'success',
            'start_time': datetime.datetime(1914, 6, 28, 10, 45, 0),
            'end_time': datetime.datetime(1914, 6, 28, 10, 45, 50),
        }
        values = shell.increment_counts(fake_test, fake_result)
        # Check to ensure counts incremented properly
        self.assertEqual(values['run_count'], 16)
        self.assertEqual(values['success'], 6)
        # Ensure run_time is updated on success
        expected_avg = ((5 * 45.0) + 50) / 6
        self.assertEqual(values['run_time'], expected_avg)

    def test_increment_counts_failure(self):
        fake_test = mock.MagicMock()
        fake_test.run_count = 15
        fake_test.success = 5
        fake_test.failure = 10
        fake_test.run_time = 45.0
        fake_result = {
            'status': 'fail',
            'start_time': datetime.datetime(1914, 6, 28, 10, 45, 0),
            'end_time': datetime.datetime(1914, 6, 28, 10, 45, 50),
        }
        values = shell.increment_counts(fake_test, fake_result)
        # Check to ensure counts incremented properly
        self.assertEqual(values['run_count'], 16)
        self.assertEqual(values['failure'], 11)
        # Avg runtime should only be updated on success
        self.assertNotIn('run_time', values)

    def test_increment_counts_skip(self):
        fake_test = mock.MagicMock()
        fake_test.run_count = 15
        fake_test.success = 5
        fake_test.failure = 10
        fake_test.run_time = 45.0
        fake_result = {
            'status': 'skip',
            'start_time': datetime.datetime(1914, 6, 28, 10, 45, 0),
            'end_time': datetime.datetime(1914, 6, 28, 10, 45, 2),
        }
        values = shell.increment_counts(fake_test, fake_result)
        # No test counts incremented with a skip
        self.assertEqual(values, {})

    def test_increment_counts_unknown_status(self):
        fake_test = mock.MagicMock()
        fake_result = {
            'status': 'explody',
            'start_time': datetime.datetime(1914, 6, 28, 10, 45, 0),
            'end_time': datetime.datetime(1914, 6, 28, 10, 45, 2),
        }
        self.assertRaises(exceptions.UnknownStatus,
                          shell.increment_counts, fake_test, fake_result)


class TestMain(base.TestCase):
    def setUp(self):
        super(TestMain, self).setUp()
        cfg.CONF.reset()
        cfg.CONF.unregister_opt(migration_cli.command_opt)
        self.fake_args = ['subunit2sql']
        self.useFixture(fixtures.MonkeyPatch('sys.argv', self.fake_args))

    @mock.patch('subunit2sql.read_subunit.ReadSubunit')
    @mock.patch('subunit2sql.shell.process_results')
    def test_main(self, process_results_mock, read_subunit_mock):
        fake_read_subunit = mock.MagicMock('ReadSubunit')
        fake_get_results = 'fake results'
        fake_read_subunit.get_results = mock.MagicMock('get_results')
        fake_read_subunit.get_results.return_value = fake_get_results
        read_subunit_mock.return_value = fake_read_subunit
        shell.main()
        read_subunit_mock.assert_called_once_with(sys.stdin,
                                                  attachments=False,
                                                  attr_regex='\[(.*)\]',
                                                  non_subunit_name=None,
                                                  targets=[],
                                                  use_wall_time=False)
        process_results_mock.assert_called_once_with(fake_get_results)

    @mock.patch('subunit2sql.read_subunit.ReadSubunit')
    @mock.patch('subunit2sql.shell.process_results')
    def test_main_multiple_files(self, process_results_mock,
                                 read_subunit_mock):
        tfile1 = tempfile.NamedTemporaryFile()
        tfile2 = tempfile.NamedTemporaryFile()
        tfile1.write(b'test me later 1')
        tfile2.write(b'test me later 2')
        tfile1.flush()
        tfile2.flush()
        self.fake_args.extend([tfile1.name, tfile2.name])
        fake_read_subunit = mock.MagicMock('ReadSubunit')
        fake_get_results_1 = 'fake results 1'
        fake_get_results_2 = 'fake results 2'
        fake_read_subunit.get_results = mock.MagicMock('get_results')
        fake_read_subunit.get_results.side_effect = [fake_get_results_1,
                                                     fake_get_results_2]
        read_subunit_mock.return_value = fake_read_subunit
        shell.main()
        read_subunit_mock.assert_called_with(mock.ANY,
                                             attachments=False,
                                             attr_regex='\[(.*)\]',
                                             non_subunit_name=None,
                                             targets=[],
                                             use_wall_time=False)
        self.assertEqual(2, len(read_subunit_mock.call_args_list))
        file_1 = read_subunit_mock.call_args_list[0][0][0]
        file_1.seek(0)
        self.assertEqual('test me later 1', file_1.read())
        file_2 = read_subunit_mock.call_args_list[1][0][0]
        file_2.seek(0)
        self.assertEqual('test me later 2', file_2.read())
        self.assertEqual(fake_get_results_1,
                         process_results_mock.call_args_list[0][0][0])
        self.assertEqual(fake_get_results_2,
                         process_results_mock.call_args_list[1][0][0])

    @mock.patch('stevedore.enabled.EnabledExtensionManager')
    @mock.patch('subunit2sql.read_subunit.ReadSubunit')
    @mock.patch('subunit2sql.shell.process_results')
    def test_main_with_targets(self, process_results_mock, read_subunit_mock,
                               ext_mock):
        exts = mock.MagicMock('EnabledExtensionManager()')
        ext_mock.return_value = exts
        exts.map = mock.MagicMock('extensions.map')
        exts.map.return_value = [mock.sentinel.extension]
        fake_read_subunit = mock.MagicMock('ReadSubunit')
        fake_get_results = 'fake results'
        fake_read_subunit.get_results = mock.MagicMock('get_results')
        fake_read_subunit.get_results.return_value = fake_get_results
        read_subunit_mock.return_value = fake_read_subunit
        shell.main()
        read_subunit_mock.assert_called_once_with(
            sys.stdin, attachments=False, attr_regex='\[(.*)\]',
            non_subunit_name=None,
            targets=[mock.sentinel.extension],
            use_wall_time=False)
        process_results_mock.assert_called_once_with(fake_get_results)


class TestProcessResults(base.TestCase):

    def setUp(self):
        super(TestProcessResults, self).setUp()
        cfg.CONF.reset()
        shell.cli_opts()
        # Ensure the config defaults are what we need
        cfg.CONF.set_override(name='run_at', override=None)
        cfg.CONF.set_override(name='artifacts', override=None)
        cfg.CONF.set_override(name='run_id', override=None)
        # Mock the whole subunit2sql.db.api module
        self.db_api_mock = self.useFixture(fixtures.MockPatch(
            'subunit2sql.shell.api')).mock
        # Mock session
        self.fake_session = mock.Mock(name='session', spec=['close'])
        self.db_api_mock.get_session.return_value = self.fake_session
        # Mock get_run_totals
        self.fake_totals = dict(skips=11, fails=22, success=33)
        self.get_run_totals_mock = self.useFixture(
            fixtures.MockPatch('subunit2sql.shell.get_run_totals',
                               return_value=self.fake_totals)).mock

    def test_process_results_no_conf(self):
        fake_run_time = 'run time'
        # Setup a common fake DB test
        fake_db_test = mock.Mock(name='db test')
        fake_db_test.id = 'test id'
        fake_db_test.run_count = 3
        fake_db_test.success = 2
        fake_db_test.failure = 1
        # Setup results
        fake_results = dict(test1={'status': 'success', 'start_time': 0,
                                   'end_time': 1, 'metadata': None,
                                   'attachments': None},
                            test2={'status': 'fail', 'start_time': 0,
                                   'end_time': 2, 'metadata': None,
                                   'attachments': None},
                            test3={'status': 'skip', 'start_time': 0,
                                   'end_time': 3, 'metadata': None,
                                   'attachments': None})

        fake_results_cpy = copy.deepcopy(fake_results)
        fake_results_cpy['run_time'] = fake_run_time
        # create mocked resources to use in testing
        fake_db_run_id = 'run id'
        fake_db_run = mock.Mock(name='db run')
        fake_db_run.id = fake_db_run_id
        self.db_api_mock.create_run.return_value = fake_db_run
        self.db_api_mock.get_test_by_test_id.return_value = fake_db_test

        # define config objects to use in our function call
        fake_run_at = '2018-04-03 00:01:00.000'
        fake_artifacts = 'artifacts'
        fake_run_id = 'run_id'
        fake_run_meta = {'run_meta': 'value'}

        # Run process_results
        shell.process_results(fake_results_cpy,
                              run_at=fake_run_at,
                              artifacts=fake_artifacts,
                              run_id=fake_run_id,
                              run_meta=fake_run_meta)
        # Check that we lookup all tests in the DB
        expected_test_by_id_calls = [mock.call(x, self.fake_session) for x in
                                     fake_results]
        self.db_api_mock.get_test_by_test_id.assert_has_calls(
            expected_test_by_id_calls, any_order=True)
        # Check that a test_run for each test is created in the DB
        expected_create_test_run_calls = [
            mock.call(fake_db_test.id, fake_db_run_id,
                      fake_results[x]['status'],
                      fake_results[x]['start_time'],
                      fake_results[x]['end_time'], self.fake_session)
            for x in fake_results]
        self.db_api_mock.create_test_run.assert_has_calls(
            expected_create_test_run_calls, any_order=True)
        self.fake_session.close.assert_called_once()

    def test_process_results_no_results(self):
        fake_run_time = 'run time'
        fake_results = dict(run_time=fake_run_time)
        fake_db_run_id = 'run id'
        fake_db_run = mock.Mock(name='db run')
        fake_db_run.id = fake_db_run_id
        self.db_api_mock.create_run.return_value = fake_db_run

        # Setup Configuration
        fake_run_at = '2016-08-17 10:58:00.000'
        cfg.CONF.set_override(name='run_at', override=fake_run_at)
        fake_artifacts = 'artifacts'
        cfg.CONF.set_override(name='artifacts', override=fake_artifacts)
        fake_run_id = 'run_id'
        cfg.CONF.set_override(name='run_id', override=fake_run_id)
        fake_run_meta = {'run_meta': 'value'}
        cfg.CONF.set_override(name='run_meta', override=fake_run_meta)
        # Run process_results
        shell.process_results(fake_results)
        self.db_api_mock.get_session.assert_called_once()
        expected_run_at = date_parser.parse(fake_run_at)
        self.db_api_mock.create_run.assert_called_once_with(
            self.fake_totals['skips'], self.fake_totals['fails'],
            self.fake_totals['success'], fake_run_time, fake_artifacts,
            id=fake_run_id, run_at=expected_run_at, session=self.fake_session)
        self.db_api_mock.add_run_metadata.assert_called_once_with(
            fake_run_meta, fake_db_run_id, self.fake_session)
        self.fake_session.close.assert_called_once()

    @mock.patch('subunit2sql.read_subunit.get_duration')
    def test_process_results_new_tests(self, get_duration_mock):
        fake_run_time = 'run time'
        fake_results = dict(test1={'status': 'success', 'start_time': 0,
                                   'end_time': 1, 'metadata': None,
                                   'attachments': None},
                            test2={'status': 'fail', 'start_time': 0,
                                   'end_time': 2, 'metadata': None,
                                   'attachments': None},
                            test3={'status': 'skip', 'start_time': 0,
                                   'end_time': 3, 'metadata': None,
                                   'attachments': None})
        fake_results_all = copy.deepcopy(fake_results)
        fake_results_all['run_time'] = fake_run_time
        # Mock create_run
        fake_db_run_id = 'run id'
        fake_db_run = mock.Mock(name='db run')
        fake_db_run.id = fake_db_run_id
        self.db_api_mock.create_run.return_value = fake_db_run
        # Tests are not found in the DB
        self.db_api_mock.get_test_by_test_id.return_value = None
        # Mock create test
        get_duration_mock.return_value = 'fake_duration'
        fake_db_test_id = 'test_db_id'
        fake_test_create = mock.Mock(name='db test')
        fake_test_create.id = fake_db_test_id
        self.db_api_mock.create_test.return_value = fake_test_create
        # Run process_results
        shell.process_results(fake_results_all)
        # Check that we lookup all tests in the DB
        expected_test_by_id_calls = [mock.call(x, self.fake_session) for x in
                                     fake_results]
        self.db_api_mock.get_test_by_test_id.assert_has_calls(
            expected_test_by_id_calls, any_order=True)
        # Check that all tests are created in the DB
        expected_create_test_calls = [
            mock.call('test1', 1, 1, 0, 'fake_duration', self.fake_session),
            mock.call('test2', 1, 0, 1, 'fake_duration', self.fake_session),
            mock.call('test3', 0, 0, 0, 'fake_duration', self.fake_session)]
        self.db_api_mock.create_test.assert_has_calls(
            expected_create_test_calls, any_order=True)
        # Check that a test_run for each test is created in the DB
        expected_create_test_run_calls = [
            mock.call(fake_db_test_id, fake_db_run_id,
                      fake_results[x]['status'], fake_results[x]['start_time'],
                      fake_results[x]['end_time'], self.fake_session)
            for x in fake_results]
        self.db_api_mock.create_test_run.assert_has_calls(
            expected_create_test_run_calls, any_order=True)
        self.fake_session.close.assert_called_once()

    @mock.patch('subunit2sql.shell.running_avg')
    def test_process_results_existing_tests(self, running_avg_mock):
        fake_run_time = 'run time'
        # Setup a common fake DB test
        fake_db_test = mock.Mock(name='db test')
        fake_db_test.id = 'test id'
        fake_db_test.run_count = 3
        fake_db_test.success = 2
        fake_db_test.failure = 1
        # Setup results
        fake_results = dict(test1={'status': 'success', 'start_time': 0,
                                   'end_time': 1, 'metadata': None,
                                   'attachments': None},
                            test2={'status': 'fail', 'start_time': 0,
                                   'end_time': 2, 'metadata': None,
                                   'attachments': None},
                            test3={'status': 'skip', 'start_time': 0,
                                   'end_time': 3, 'metadata': None,
                                   'attachments': None})
        fake_results_all = copy.deepcopy(fake_results)
        fake_results_all['run_time'] = fake_run_time
        # Mock create run
        fake_db_run_id = 'run id'
        fake_db_run = mock.Mock(name='db run')
        fake_db_run.id = fake_db_run_id
        self.db_api_mock.create_run.return_value = fake_db_run
        # Tests are found in the DB
        self.db_api_mock.get_test_by_test_id.return_value = fake_db_test
        # Mock running avg
        fake_running_avg = 'running average'

        def fake_running_avg_method(test, values, result):
            values['run_time'] = fake_running_avg
            return values

        running_avg_mock.side_effect = fake_running_avg_method

        # Run process_results
        shell.process_results(fake_results_all)
        # Check that we lookup all tests in the DB
        expected_test_by_id_calls = [mock.call(x, self.fake_session) for x in
                                     fake_results]
        self.db_api_mock.get_test_by_test_id.assert_has_calls(
            expected_test_by_id_calls, any_order=True)
        # Check that counters for tests are updated (if not skip)
        expected_update_test_calls = [
            mock.call({'run_count': 4, 'success': 3,
                       'run_time': fake_running_avg},
                      fake_db_test.id, self.fake_session),
            mock.call({'run_count': 4, 'failure': 2},
                      fake_db_test.id, self.fake_session)]
        self.db_api_mock.update_test.assert_has_calls(
            expected_update_test_calls, any_order=True)
        # Check that a test_run for each test is created in the DB
        expected_create_test_run_calls = [
            mock.call(fake_db_test.id, fake_db_run_id,
                      fake_results[x]['status'],
                      fake_results[x]['start_time'],
                      fake_results[x]['end_time'], self.fake_session)
            for x in fake_results]
        self.db_api_mock.create_test_run.assert_has_calls(
            expected_create_test_run_calls, any_order=True)
        self.fake_session.close.assert_called_once()

    def test_process_results_existing_tests_invalid_status(self):
        fake_run_time = 'run time'
        # Setup a common fake DB test
        fake_db_test = mock.Mock(name='db test')
        fake_db_test.run_count = 3
        # Setup results
        fake_results = dict(test1={'status': 'invalid', 'start_time': 0,
                                   'end_time': 1, 'metadata': None,
                                   'attachments': None})
        fake_results_all = copy.deepcopy(fake_results)
        fake_results_all['run_time'] = fake_run_time
        # Tests are found in the DB
        self.db_api_mock.get_test_by_test_id.return_value = fake_db_test
        with testtools.ExpectedException(exceptions.UnknownStatus,
                                         '^.*\n.*%s$' % 'invalid'):
            # Run process_results
            shell.process_results(fake_results_all)
        # Check that we lookup all tests in the DB
        expected_test_by_id_calls = [mock.call(x, self.fake_session) for x in
                                     fake_results]
        self.db_api_mock.get_test_by_test_id.assert_has_calls(
            expected_test_by_id_calls, any_order=True)
        # Check that a run is created in the DB
        self.db_api_mock.create_run.assert_called_once_with(
            self.fake_totals['skips'], self.fake_totals['fails'],
            self.fake_totals['success'], fake_run_time, None, id=None,
            run_at=None, session=self.fake_session)
        # Check no test run was added to the DB
        self.db_api_mock.create_test_run.assert_not_called()
        # FIXME(andreaf) Session is not closed.
        # https://storyboard.openstack.org/?#!/story/2000702
        # self.fake_session.close.assert_called_once()

    def test_process_results_test_metadata_new(self):
        # Setup the metadata prefix configuration
        fake_meta_prefix = '_meta_'
        cfg.CONF.set_override(name='test_attr_prefix',
                              override=fake_meta_prefix)
        fake_run_time = 'run time'
        # Setup a common fake DB test
        fake_db_test = mock.Mock(name='db test')
        fake_db_test.id = 'test id'
        fake_db_test.run_count = 3
        fake_db_test.success = 2
        fake_db_test.failure = 1
        # Setup results
        expected_metadata_list = [fake_meta_prefix + 'a',
                                  fake_meta_prefix + 'b']
        extra_metadata_list = ['c', 'd']
        full_metadata_list = expected_metadata_list + extra_metadata_list
        fake_results = dict(
            test1={'status': 'fail', 'start_time': 0, 'end_time': 1,
                   'metadata': {'attrs': ','.join(full_metadata_list)},
                   'attachments': None})
        fake_results_all = copy.deepcopy(fake_results)
        fake_results_all['run_time'] = fake_run_time
        # Mock create run
        fake_db_run_id = 'run id'
        fake_db_run = mock.Mock(name='db run')
        fake_db_run.id = fake_db_run_id
        self.db_api_mock.create_run.return_value = fake_db_run
        # Mock create test run
        fake_db_test_run_id = 'test run id'
        fake_db_test_run = mock.Mock(name='db test run')
        fake_db_test_run.id = fake_db_test_run_id
        self.db_api_mock.create_test_run.return_value = fake_db_test_run
        # Tests are found in the DB
        self.db_api_mock.get_test_by_test_id.return_value = fake_db_test
        # Test metadata is not found in the DB
        self.db_api_mock.get_test_metadata.return_value = []
        # Run process_results
        shell.process_results(fake_results_all)
        # Check only matching metadata is added to test_metadata
        expected_add_test_metadata_calls = [
            mock.call({'attr': x}, fake_db_test.id, session=self.fake_session)
            for x in expected_metadata_list]
        self.db_api_mock.add_test_metadata.assert_has_calls(
            expected_add_test_metadata_calls, any_order=True)
        # Check all metadata is added to test_run_metadata
        self.db_api_mock.add_test_run_metadata.assert_has_calls([
            mock.call(fake_results['test1']['metadata'], fake_db_test_run_id,
                      self.fake_session)])
        self.fake_session.close.assert_called_once()

    def test_process_results_test_metadata_existing(self):
        # Setup the metadata prefix configuration
        fake_meta_prefix = '_meta_'
        cfg.CONF.set_override(name='test_attr_prefix',
                              override=fake_meta_prefix)
        fake_run_time = 'run time'
        # Setup a common fake DB test
        fake_db_test = mock.Mock(name='db test')
        fake_db_test.id = 'test id'
        fake_db_test.run_count = 3
        fake_db_test.success = 2
        fake_db_test.failure = 1
        # Setup results
        expected_metadata_list = [fake_meta_prefix + 'a',
                                  fake_meta_prefix + 'b']
        existing_metadata_list = [fake_meta_prefix + 'c',
                                  fake_meta_prefix + 'd']
        full_metadata_list = expected_metadata_list + existing_metadata_list
        fake_results = dict(
            test1={'status': 'fail', 'start_time': 0, 'end_time': 1,
                   'metadata': {'attrs': ','.join(full_metadata_list)},
                   'attachments': None})
        fake_results_all = copy.deepcopy(fake_results)
        fake_results_all['run_time'] = fake_run_time
        # Mock create run
        fake_db_run_id = 'run id'
        fake_db_run = mock.Mock(name='db run')
        fake_db_run.id = fake_db_run_id
        self.db_api_mock.create_run.return_value = fake_db_run
        # Mock create test run
        fake_db_test_run_id = 'test run id'
        fake_db_test_run = mock.Mock(name='db test run')
        fake_db_test_run.id = fake_db_test_run_id
        self.db_api_mock.create_test_run.return_value = fake_db_test_run
        # Tests are found in the DB
        self.db_api_mock.get_test_by_test_id.return_value = fake_db_test
        # Test metadata is found in the DB
        test_metadata_list = []
        for value in existing_metadata_list:
            test_metadata = mock.Mock()
            test_metadata.key = 'attr'
            test_metadata.value = value
            test_metadata_list.append(test_metadata)
        self.db_api_mock.get_test_metadata.return_value = test_metadata_list
        # Run process_results
        shell.process_results(fake_results_all)
        # Check only matching metadata is added to test_metadata
        expected_add_test_metadata_calls = [
            mock.call({'attr': x}, fake_db_test.id, session=self.fake_session)
            for x in expected_metadata_list]
        self.db_api_mock.add_test_metadata.assert_has_calls(
            expected_add_test_metadata_calls, any_order=True)
        # Check all metadata is added to test_run_metadata
        self.db_api_mock.add_test_run_metadata.assert_has_calls([
            mock.call(fake_results['test1']['metadata'], fake_db_test_run_id,
                      self.fake_session)])
        self.fake_session.close.assert_called_once()

    def test_process_result_test_metadata_no_prefix(self):
        # Setup the metadata prefix configuration
        cfg.CONF.set_override(name='test_attr_prefix',
                              override='')
        fake_run_time = 'run time'
        # Setup a common fake DB test
        fake_db_test = mock.Mock(name='db test')
        fake_db_test.id = 'test id'
        fake_db_test.run_count = 3
        fake_db_test.success = 2
        fake_db_test.failure = 1
        # Setup results
        expected_metadata_list = ['a', 'b', 'c']
        fake_results = dict(
            test1={'status': 'fail', 'start_time': 0, 'end_time': 1,
                   'metadata': {'attrs': ','.join(expected_metadata_list)},
                   'attachments': None})
        fake_results_all = copy.deepcopy(fake_results)
        fake_results_all['run_time'] = fake_run_time
        # Mock create run
        fake_db_run_id = 'run id'
        fake_db_run = mock.Mock(name='db run')
        fake_db_run.id = fake_db_run_id
        self.db_api_mock.create_run.return_value = fake_db_run
        # Mock create test run
        fake_db_test_run_id = 'test run id'
        fake_db_test_run = mock.Mock(name='db test run')
        fake_db_test_run.id = fake_db_test_run_id
        self.db_api_mock.create_test_run.return_value = fake_db_test_run
        # Tests are found in the DB
        self.db_api_mock.get_test_by_test_id.return_value = fake_db_test
        # Run process_results
        shell.process_results(fake_results_all)
        # Check test metadata is not added
        self.db_api_mock.add_test_metadata.assert_not_called()
        # Check all metadata is added to test_run_metadata
        self.db_api_mock.add_test_run_metadata.assert_has_calls([
            mock.call(fake_results['test1']['metadata'],
                      fake_db_test_run_id,
                      self.fake_session)])
        self.fake_session.close.assert_called_once()

    def test_process_result_test_run_attachments(self):
        fake_run_time = 'run time'
        # Setup a common fake DB test
        fake_db_test = mock.Mock(name='db test')
        fake_db_test.id = 'test id'
        fake_db_test.run_count = 3
        fake_db_test.success = 2
        fake_db_test.failure = 1
        # Setup results
        fake_attachment = 'some text'
        fake_results = dict(
            test1={'status': 'fail', 'start_time': 0, 'end_time': 1,
                   'metadata': None, 'attachments': fake_attachment})
        fake_results_all = copy.deepcopy(fake_results)
        fake_results_all['run_time'] = fake_run_time
        # Mock create run
        fake_db_run_id = 'run id'
        fake_db_run = mock.Mock(name='db run')
        fake_db_run.id = fake_db_run_id
        self.db_api_mock.create_run.return_value = fake_db_run
        # Mock create test run
        fake_db_test_run_id = 'test run id'
        fake_db_test_run = mock.Mock(name='db test run')
        fake_db_test_run.id = fake_db_test_run_id
        self.db_api_mock.create_test_run.return_value = fake_db_test_run
        # Tests are found in the DB
        self.db_api_mock.get_test_by_test_id.return_value = fake_db_test
        # Run process_results
        shell.process_results(fake_results_all)
        # Check attachments are added to the test run
        self.db_api_mock.add_test_metadata.assert_not_called()
        # Check all metadata is added to test_run_metadata
        self.db_api_mock.add_test_run_attachments.assert_has_calls([
            mock.call(fake_attachment, fake_db_test_run_id,
                      self.fake_session)])
        self.fake_session.close.assert_called_once()

    def test_process_results_test_metadata_new_remove_prefix(self):
        # Setup the metadata prefix configuration
        fake_meta_prefix = '_meta_'
        cfg.CONF.set_override(name='test_attr_prefix',
                              override=fake_meta_prefix)
        cfg.CONF.set_override(name='remove_test_attr_prefix',
                              override=True)
        fake_run_time = 'run time'
        # Setup a common fake DB test
        fake_db_test = mock.Mock(name='db test')
        fake_db_test.id = 'test id'
        fake_db_test.run_count = 3
        fake_db_test.success = 2
        fake_db_test.failure = 1
        # Setup results
        expected_metadata_list = ['a', 'b']
        extra_metadata_list = ['c', 'd']
        full_metadata_list = [fake_meta_prefix + x for x in
                              expected_metadata_list] + extra_metadata_list
        fake_results = dict(
            test1={'status': 'fail', 'start_time': 0, 'end_time': 1,
                   'metadata': {'attrs': ','.join(full_metadata_list)},
                   'attachments': None})
        fake_results_all = copy.deepcopy(fake_results)
        fake_results_all['run_time'] = fake_run_time
        # Mock create run
        fake_db_run_id = 'run id'
        fake_db_run = mock.Mock(name='db run')
        fake_db_run.id = fake_db_run_id
        self.db_api_mock.create_run.return_value = fake_db_run
        # Mock create test run
        fake_db_test_run_id = 'test run id'
        fake_db_test_run = mock.Mock(name='db test run')
        fake_db_test_run.id = fake_db_test_run_id
        self.db_api_mock.create_test_run.return_value = fake_db_test_run
        # Tests are found in the DB
        self.db_api_mock.get_test_by_test_id.return_value = fake_db_test
        # Test metadata is not found in the DB
        self.db_api_mock.get_test_metadata.return_value = []
        # Run process_results
        shell.process_results(fake_results_all)
        # Check only matching metadata is added to test_metadata
        expected_add_test_metadata_calls = [
            mock.call({'attr': x}, fake_db_test.id, session=self.fake_session)
            for x in expected_metadata_list]
        self.db_api_mock.add_test_metadata.assert_has_calls(
            expected_add_test_metadata_calls, any_order=True)
        # Check all metadata is added to test_run_metadata
        self.db_api_mock.add_test_run_metadata.assert_has_calls([
            mock.call(fake_results['test1']['metadata'], fake_db_test_run_id,
                      self.fake_session)])
        self.fake_session.close.assert_called_once()

    def test_process_results_test_metadata_existing_remove_prefix(self):
        # Setup the metadata prefix configuration
        fake_meta_prefix = '_meta_'
        cfg.CONF.set_override(name='test_attr_prefix',
                              override=fake_meta_prefix)
        cfg.CONF.set_override(name='remove_test_attr_prefix',
                              override=True)
        fake_run_time = 'run time'
        # Setup a common fake DB test
        fake_db_test = mock.Mock(name='db test')
        fake_db_test.id = 'test id'
        fake_db_test.run_count = 3
        fake_db_test.success = 2
        fake_db_test.failure = 1
        # Setup results
        expected_metadata_list = ['a', 'b']
        existing_metadata_list = ['c', 'd']
        full_metadata_list = [fake_meta_prefix + x for x in (
            expected_metadata_list + existing_metadata_list)]
        fake_results = dict(
            test1={'status': 'fail', 'start_time': 0, 'end_time': 1,
                   'metadata': {'attrs': ','.join(full_metadata_list)},
                   'attachments': None})
        fake_results_all = copy.deepcopy(fake_results)
        fake_results_all['run_time'] = fake_run_time
        # Mock create run
        fake_db_run_id = 'run id'
        fake_db_run = mock.Mock(name='db run')
        fake_db_run.id = fake_db_run_id
        self.db_api_mock.create_run.return_value = fake_db_run
        # Mock create test run
        fake_db_test_run_id = 'test run id'
        fake_db_test_run = mock.Mock(name='db test run')
        fake_db_test_run.id = fake_db_test_run_id
        self.db_api_mock.create_test_run.return_value = fake_db_test_run
        # Tests are found in the DB
        self.db_api_mock.get_test_by_test_id.return_value = fake_db_test
        # Test metadata is found in the DB
        test_metadata_list = []
        for value in existing_metadata_list:
            test_metadata = mock.Mock()
            test_metadata.key = 'attr'
            test_metadata.value = value
            test_metadata_list.append(test_metadata)
        self.db_api_mock.get_test_metadata.return_value = test_metadata_list
        # Run process_results
        shell.process_results(fake_results_all)
        # Check only matching metadata is added to test_metadata
        expected_add_test_metadata_calls = [
            mock.call({'attr': x}, fake_db_test.id, session=self.fake_session)
            for x in expected_metadata_list]
        self.db_api_mock.add_test_metadata.assert_has_calls(
            expected_add_test_metadata_calls, any_order=True)
        # Check all metadata is added to test_run_metadata
        self.db_api_mock.add_test_run_metadata.assert_has_calls([
            mock.call(fake_results['test1']['metadata'], fake_db_test_run_id,
                      self.fake_session)])
        self.fake_session.close.assert_called_once()



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import tempfile

from alembic import command
from alembic import config as alembic_config
from oslo_config import cfg
import sqlalchemy

from subunit2sql.db import api as session

CONF = cfg.CONF
SQLITE_TEST_DATABASE_PATH = tempfile.mkstemp('subunit2sql.db')[1]

script_location = os.path.join(os.path.dirname(os.path.dirname(
    os.path.abspath(__file__))), 'migrations')


def get_connect_string(backend,
                       user="openstack_citest",
                       passwd="openstack_citest",
                       database="openstack_citest"):
    """Generate a db uri for testing locally.

    Try to get a connection with a very specific set of values, if we get
    these then we'll run the tests, otherwise they are skipped
    """
    if backend == "mysql":
        backend = "mysql+pymysql"
    elif backend == "postgres":
        backend = "postgresql+psycopg2"

    if backend == "sqlite":
        return "sqlite:///" + SQLITE_TEST_DATABASE_PATH

    return ("%(backend)s://%(user)s:%(passwd)s@localhost/%(database)s"
            % {'backend': backend, 'user': user, 'passwd': passwd,
               'database': database})


def is_backend_avail(backend,
                     user="openstack_citest",
                     passwd="openstack_citest",
                     database="openstack_citest"):
    try:
        if backend == "mysql":
            connect_uri = get_connect_string("mysql", user=user,
                                             passwd=passwd, database=database)
        elif backend == "postgres":
            connect_uri = get_connect_string("postgres", user=user,
                                             passwd=passwd, database=database)
        elif backend == "sqlite":
            return True
        engine = sqlalchemy.create_engine(connect_uri)
        connection = engine.connect()
    except Exception:
        # intentionally catch all to handle exceptions even if we don't
        # have any backend code loaded.
        return False
    else:
        connection.close()
        engine.dispose()
        return True


def run_migration(target, engine=None):
    engine = engine or session.get_engine()
    engine.connect()
    config = alembic_config.Config(os.path.join(script_location,
                                                'alembic.ini'))
    config.set_main_option('script_location', 'subunit2sql:migrations')
    config.subunit2sql_config = CONF
    with engine.begin() as connection:
        config.attributes['connection'] = connection
        command.upgrade(config, target)
    engine.dispose()



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Cleanup and Improve Indexes

Revision ID: b96122f780
Revises: 2fb76f1a1393
Create Date: 2015-10-14 12:03:26.965724

"""

# revision identifiers, used by Alembic.
revision = 'b96122f780'
down_revision = '2fb76f1a1393'

from alembic import context
from alembic import op
from sqlalchemy.engine import reflection


def upgrade():
    migration_context = context.get_context()
    insp = reflection.Inspector(migration_context.bind)
    test_run_indx = insp.get_indexes('test_runs')
    test_run_indx_names = [x['name'] for x in test_run_indx]
    # Cleanup any duplicate indexes on test_runs
    if 'ix_test_runs_test_id' in test_run_indx_names:
        if 'ix_test_run_test_id' in test_run_indx_names:
            op.drop_index('ix_test_run_test_id', 'test_runs')
    if 'ix_test_runs_run_id' in test_run_indx_names:
        if 'ix_test_run_run_id' in test_run_indx_names:
            op.drop_index('ix_test_run_run_id', 'test_runs')

    # Add an index for test, test_id
    op.create_index('ix_test_ids', 'tests', ['id', 'test_id'], mysql_length=30)


def downgrade():
    pass



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import io

import mock
import subunit as subunit_lib

from subunit2sql import read_subunit as subunit
from subunit2sql.tests import base


class TestReadSubunit(base.TestCase):
    def test_get_duration(self):
        dur = subunit.get_duration(datetime.datetime(1914, 6, 28, 10, 45, 0),
                                   datetime.datetime(1914, 6, 28, 10, 45, 50))
        self.assertEqual(dur, 50.000000)

    def test_get_duration_no_start(self):
        dur = subunit.get_duration(None,
                                   datetime.datetime(1914, 6, 28, 10, 45, 50))
        self.assertIsNone(dur)

    def test_get_duration_no_end(self):
        dur = subunit.get_duration(datetime.datetime(1914, 6, 28, 10, 45, 50),
                                   None)
        self.assertIsNone(dur)

    def test_get_attrs(self):
        fake_subunit = subunit.ReadSubunit(mock.MagicMock())
        fake_name = 'fake_dir.fake_module.fakeClass.test_fake[attr1,attr2]'
        attrs = fake_subunit.get_attrs(fake_name)
        self.assertEqual(attrs, 'attr1,attr2')

    def test_cleanup_test_name_with_attrs(self):
        fake_subunit = subunit.ReadSubunit(mock.MagicMock())
        fake_name = 'fake_dir.fake_module.fakeClass.test_fake[attr1,attr2]'
        name = fake_subunit.cleanup_test_name(fake_name)
        self.assertEqual(name, 'fake_dir.fake_module.fakeClass.test_fake')

    def test_cleanup_test_name_with_attrs_leave_scenarios(self):
        fake_subunit = subunit.ReadSubunit(mock.MagicMock())
        fake_name = ('fake_dir.fake_module.fakeClass.test_fake[attr1,attr2]'
                     '(scenario)')
        name = fake_subunit.cleanup_test_name(fake_name)
        self.assertEqual(name, 'fake_dir.fake_module.fakeClass.test_fake'
                         '(scenario)')

    def test_cleanup_test_name_with_strip_scenarios_and_attrs(self):
        fake_subunit = subunit.ReadSubunit(mock.MagicMock())
        fake_name = ('fake_dir.fake_module.fakeClass.test_fake[attr1,attr2]'
                     '(scenario)')
        name = fake_subunit.cleanup_test_name(fake_name, strip_scenarios=True)
        self.assertEqual(name, 'fake_dir.fake_module.fakeClass.test_fake')

    def test_cleanup_test_name_strip_nothing(self):
        fake_subunit = subunit.ReadSubunit(mock.MagicMock())
        fake_name = ('fake_dir.fake_module.fakeClass.test_fake[attr1,attr2]'
                     '(scenario)')
        name = fake_subunit.cleanup_test_name(fake_name, strip_tags=False)
        self.assertEqual(name, fake_name)

    def test_run_time(self):
        fake_subunit = subunit.ReadSubunit(mock.MagicMock())
        fake_results = {}
        fifty_sec_run_result = {
            'start_time': datetime.datetime(1914, 6, 28, 10, 45, 0),
            'end_time': datetime.datetime(1914, 6, 28, 10, 45, 50),
        }
        for num in range(100):
            test_name = 'test_fake_' + str(num)
            fake_results[test_name] = fifty_sec_run_result
        fake_subunit.results = fake_results
        runtime = fake_subunit.run_time()
        self.assertEqual(runtime, 5000.0)

    def test_wall_run_time(self):
        fake_subunit = subunit.ReadSubunit(mock.MagicMock(),
                                           use_wall_time=True)
        fake_results = {}
        start_time = datetime.datetime(1914, 6, 28, 10, 45, 0)
        stop_time = datetime.datetime(1914, 6, 28, 10, 45, 50)
        fifty_sec_run_result = {
            'start_time': start_time,
            'end_time': stop_time,
        }
        fake_results['first'] = fifty_sec_run_result
        for num in range(100):
            test_name = 'test_fake_' + str(num)
            start_time = start_time + datetime.timedelta(minutes=1)
            stop_time = stop_time + datetime.timedelta(minutes=1)
            fake_result = {
                'start_time': start_time,
                'end_time': stop_time,
            }
            fake_results[test_name] = fake_result
        fake_subunit.results = fake_results
        runtime = fake_subunit.run_time()
        # Wall time should be (60 * 100) + 50
        self.assertEqual(runtime, 6050.0)

    def test_parse_outcome(self):
        fake_subunit = subunit.ReadSubunit(mock.MagicMock())

        fake_id = 'fake_dir.fake_module.fakeClass.test_fake[attr1,attr2]'
        fake_timestamps = [datetime.datetime(1914, 8, 26, 20, 00, 00),
                           datetime.datetime(2014, 8, 26, 20, 00, 00)]
        fake_status = 'skip'
        tags = set(['worker-0'])

        fake_results = {
            'status': fake_status,
            'details': {
                'reason': 'fake reason'
            },
            'id': fake_id,
            'timestamps': fake_timestamps,
            'tags': tags
        }
        fake_subunit.parse_outcome(fake_results)

        parsed_results = fake_subunit.results
        # assert that the dict root key is the test name - the fake_id stripped
        # of the tags
        fake_test_name = fake_id[:fake_id.find('[')]
        self.assertEqual(list(parsed_results.keys()), [fake_test_name])

        self.assertEqual(parsed_results[fake_test_name]['status'],
                         fake_status)
        self.assertEqual(parsed_results[fake_test_name]['start_time'],
                         fake_timestamps[0])
        self.assertEqual(parsed_results[fake_test_name]['end_time'],
                         fake_timestamps[1])
        fake_attrs = fake_id[fake_id.find('[') + 1:fake_id.find(']')]
        self.assertEqual(parsed_results[fake_test_name]['metadata']['attrs'],
                         fake_attrs)
        self.assertEqual(parsed_results[fake_test_name]['metadata']['tags'],
                         ','.join(tags))

    def test_attrs_default_regex(self):
        fake_id = 'test_fake.TestThing.testA[good_test,fun_test,legit]'
        read = subunit.ReadSubunit(io.BytesIO())
        attrs = read.get_attrs(fake_id)
        attrs_list = attrs.split(',')
        self.assertIn('legit', attrs_list)
        self.assertIn('fun_test', attrs_list)
        self.assertIn('good_test', attrs_list)

    def test_attrs_non_default_regex(self):
        fake_id = 'test_fake.TestThing.testB`good_test,fun_test,legit`'
        regex = '\`(.*)\`'
        read = subunit.ReadSubunit(io.BytesIO(), attr_regex=regex)
        attrs = read.get_attrs(fake_id)
        attrs_list = attrs.split(',')
        self.assertIn('legit', attrs_list)
        self.assertIn('fun_test', attrs_list)
        self.assertIn('good_test', attrs_list)

    def test_attrs_no_matches(self):
        fake_id = 'test_fake.TestThing.testB`good_test,fun_test,legit`'
        read = subunit.ReadSubunit(io.BytesIO())
        attrs = read.get_attrs(fake_id)
        self.assertIsNone(attrs)

    def test_cleanup_test_name_default_attr_regex(self):
        fake_id = 'test_fake.TestThing.testA[good_test,fun_test,legit]'
        read = subunit.ReadSubunit(io.BytesIO())
        test_name = read.cleanup_test_name(fake_id, strip_tags=True,
                                           strip_scenarios=False)
        self.assertEqual('test_fake.TestThing.testA', test_name)

    def test_cleanup_test_name_non_default_attr_regex(self):
        fake_id = 'test_fake.TestThing.testB`good_test,fun_test,legit`'
        regex = '\`(.*)\`'
        read = subunit.ReadSubunit(io.BytesIO(), attr_regex=regex)
        test_name = read.cleanup_test_name(fake_id, strip_tags=True,
                                           strip_scenarios=False)
        self.assertEqual('test_fake.TestThing.testB', test_name)

    def test_cleanup_test_name_no_attr_matches(self):
        fake_id = 'test_fake.TestThing.testB`good_test,fun_test,legit`'
        read = subunit.ReadSubunit(io.BytesIO())
        test_name = read.cleanup_test_name(fake_id, strip_tags=True,
                                           strip_scenarios=False)
        self.assertEqual(fake_id, test_name)

    @mock.patch('testtools.CopyStreamResult')
    def test_targets_added_to_result(self, ttc_mock):
        subunit.ReadSubunit(mock.MagicMock(), targets=['foo'])
        self.assertIn('foo', ttc_mock.call_args[0][0])

    @mock.patch('testtools.CopyStreamResult')
    def test_targets_not_modified(self, ttc_mock):
        # A regression test that verifies that the subunit reader does
        # not modify the passed in value of targets.
        targets = ['foo']

        subunit.ReadSubunit(mock.MagicMock(), targets=targets)
        ntargets1 = len(ttc_mock.call_args[0][0])

        subunit.ReadSubunit(mock.MagicMock(), targets=targets)
        ntargets2 = len(ttc_mock.call_args[0][0])

        self.assertEqual(ntargets1, ntargets2)
        self.assertEqual(targets, ['foo'])

    def test_non_subunit_name(self):
        non_subunit_name = 'fake_non_subunit'
        fake_subunit = subunit.ReadSubunit(mock.MagicMock(),
                                           non_subunit_name=non_subunit_name)
        self.assertEqual(fake_subunit.stream.non_subunit_name,
                         non_subunit_name)

    def test_not_subunit_no_subunit_name_set(self):
        stream_buf = io.BytesIO()
        stream = subunit_lib.StreamResultToBytes(stream_buf)
        stream.status(test_id='test_a', test_status='inprogress')
        stream.status(test_id='test_a', test_status='success')
        stream_buf.write(b'I AM NOT SUBUNIT')
        stream_buf.seek(0)
        result = subunit.ReadSubunit(stream_buf)
        exc_found = False
        try:
            result.get_results()
        # NOTE(mtreinish): Subunit raises the generic Exception class
        # so manually inspect the Exception object to check the error
        # message
        except Exception as e:
            self.assertIsInstance(e, Exception)
            self.assertEqual(e.args, ('Non subunit content', b'I'))
            exc_found = True
        self.assertTrue(exc_found,
                        'subunit exception not raised on invalid content')



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import sys

from oslo_config import cfg
import stevedore

import subunit2sql.analysis.agg_count
import subunit2sql.analysis.ci_time
import subunit2sql.analysis.dailycount
import subunit2sql.analysis.failures
import subunit2sql.analysis.run_failure_rate
import subunit2sql.analysis.run_time
import subunit2sql.analysis.run_time_meta
import subunit2sql.analysis.test_run_time
from subunit2sql import shell

CONF = cfg.CONF
CONF.import_opt('verbose', 'subunit2sql.db.api')

SHELL_OPTS = [
    cfg.StrOpt('title', short='t', help='Optional title to use for the graph '
                                        'output. If one is not specified, the '
                                        'full test_id will be used'),
    cfg.StrOpt('output', short='o', required=True,
               help='Output path to write image file to. The file extension '
                    'will determine the file format.'),
    cfg.StrOpt('start-date', short='d',
               help='Start date for the graph only data from after this date '
                    'will be used. Uses ISO 8601 format: 1914-06-28'),
    cfg.StrOpt('stop-date', short='s',
               help='Stop date for the graph only data from before this date '
                    'will be used. Uses ISO 8601 format: 1914-06-28'),
    cfg.IntOpt('dpi', default=900,
               help='Image DPI to use for the output graph images')
]


def get_plugin_list():
    plugin_list = stevedore.ExtensionManager(
        'subunit2sql.graph.plugin',
        invoke_on_load=True,
        propagate_map_exceptions=True)
    return plugin_list


def add_command_parsers(subparsers):
    graph_commands = {}
    # Put commands from in-tree commands on init list
    for command in ['failures', 'run_time', 'agg_count', 'dailycount',
                    'run_failure_rate', 'run_time_meta', 'test_run_time',
                    'ci_time']:
        graph_commands[command] = getattr(subunit2sql.analysis, command)

    # Load any installed out of tree commands on the init list
    for plugin in get_plugin_list():
        graph_commands[plugin.name] = plugin.plugin

    # Init all commands from graph_commands
    for name in graph_commands:
        parser = subparsers.add_parser(name)
        graph_commands[name].set_cli_opts(parser)
        parser.set_defaults(
            func=graph_commands[name].generate_series)

command_opt = cfg.SubCommandOpt('command', title='graph',
                                help='Available graphs',
                                handler=add_command_parsers)


def cli_opts():
    for opt in SHELL_OPTS:
        CONF.register_cli_opt(opt)
    CONF.register_cli_opt(command_opt)


def list_opts():
    opt_list = copy.deepcopy(SHELL_OPTS)
    return [('DEFAULT', opt_list)]


def main():
    cli_opts()
    shell.parse_args(sys.argv)
    CONF.command.func()
    print('Graph saved at: %s' % CONF.output)


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""create missing indexes

Revision ID: 13d819bbb0ff
Revises: 4ca26dac400e
Create Date: 2014-06-20 21:27:01.629724

"""

# revision identifiers, used by Alembic.
revision = '13d819bbb0ff'
down_revision = '4ca26dac400e'

from alembic import context
from alembic import op
from sqlalchemy.engine import reflection


def upgrade():
    migration_context = context.get_context()
    insp = reflection.Inspector(migration_context.bind)
    test_indx = insp.get_indexes('tests')
    test_indx_names = [x['name'] for x in test_indx]
    test_indx_columns = [x['column_names'][0] for x in test_indx
                         if len(x) == 1]
    test_run_indx = insp.get_indexes('test_runs')
    test_run_indx_names = [x['name'] for x in test_run_indx]
    test_run_indx_columns = [x['column_names'][0] for x in test_run_indx
                             if len(x) == 1]
    if ('ix_test_id' not in test_indx_names and
        'test_id' not in test_indx_columns):
        op.create_index('ix_test_id', 'tests', ['test_id'], mysql_length=30)

    # remove auto created indexes (sqlite only)
    # note the name is with test_runs not test_run
    if migration_context.dialect.name == 'sqlite':
        if 'ix_test_runs_test_id' in test_run_indx_names:
            op.drop_index('ix_test_runs_test_id', 'test_runs')
        if 'ix_test_runs_run_id' in test_run_indx_names:
            op.drop_index('ix_test_runs_run_id', 'test_runs')

    with op.batch_alter_table('test_runs') as batch_op:
        batch_op.create_unique_constraint('uq_test_runs',
                                          ['test_id', 'run_id'])

    if ('ix_test_run_test_id' not in test_run_indx_names and
        'test_id' not in test_run_indx_columns):
        op.create_index('ix_test_run_test_id', 'test_runs', ['test_id'])
    if ('ix_test_run_run_id' not in test_run_indx_names and
        'run_id' not in test_run_indx_columns):
        op.create_index('ix_test_run_run_id', 'test_runs', ['run_id'])


def downgrade():
    op.drop_constraint('uq_test_runs', 'test_runs')
    op.drop_index('ix_test_id', 'tests')
    op.drop_index('ix_test_run_test_id', 'test_runs')
    op.drop_index('ix_test_run_run_id', 'test_runs')



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import matplotlib
import matplotlib.dates as dates
import matplotlib.pyplot as plt
from oslo_config import cfg
import pandas as pd

from subunit2sql.analysis import utils
from subunit2sql.db import api

CONF = cfg.CONF

matplotlib.style.use('ggplot')


def set_cli_opts(parser):
    parser.add_argument('test_id', nargs='?',
                        help='Test id to extract time series for. This '
                             'previously took a UUID from the tests.id '
                             'column, however this will no longer work. It '
                             'only works with a value from tests.test_id.')


def generate_series():
    session = api.get_session()
    test_id = api.get_id_from_test_id(CONF.command.test_id, session)
    if not test_id:
        print("The test_id %s was not found in the database" %
              CONF.command.test_id)
        exit(2)
    run_times = api.get_test_run_time_series(test_id, session)
    if not run_times:
        print("There was no data found in the database")
        exit(3)
    if not CONF.title:
        test = api.get_test_by_id(test_id, session)
    session.close()
    ts = pd.Series(run_times)
    ts = utils.filter_dates(ts)
    if ts.count() == 0:
        print("No data available. Check your query and try again.")
        exit(-1)
    roll = ts.rolling(window=20, center=False)
    mean = roll.mean()
    rolling_std = roll.std()
    plt.figure()
    if not CONF.title:
        plt.title(test.test_id)
    else:
        plt.title(CONF.title)
    plt.ylabel('Time (sec.)')

    # format x-axis with dates
    fig, ax = plt.subplots(1)
    fig.autofmt_xdate()
    xfmt = dates.DateFormatter("%b %d %Y")
    ax.xaxis_date()
    ax.xaxis.set_major_formatter(xfmt)

    plt.plot(ts.index, ts, 'ko', label='Run Time', markersize=0.45)
    plt.plot(mean.index, mean, 'b', label='Avg. Run Time', linewidth=0.45)
    upper_std_dev = mean + 2 * rolling_std
    lower_std_dev = mean - 2 * rolling_std
    # Set negative numbers to 0
    lower_std_dev[lower_std_dev < 0] = 0
    plt.fill_between(rolling_std.index, upper_std_dev,
                     lower_std_dev, color='b', alpha=0.2,
                     label='std dev')
    plt.legend()
    plt.savefig(CONF.output, dpi=CONF.dpi)
    return ts



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""create tests tables

Revision ID: 5ef013efbc2
Revises: None
Create Date: 2014-06-08 11:18:41.529268

"""

# revision identifiers, used by Alembic.
revision = '5ef013efbc2'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('tests',
                    sa.Column('id', sa.String(36), primary_key=True,
                              nullable=False),
                    sa.Column('test_id', sa.String(256), nullable=False),
                    sa.Column('run_count', sa.Integer()),
                    sa.Column('success', sa.Integer()),
                    sa.Column('failure', sa.Integer()),
                    mysql_engine='InnoDB')


def downgrade():
    op.drop_table('tables')



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Rename test_id column in test_metadata table

Revision ID: 2fb76f1a1393
Revises: 1ff737bef438
Create Date: 2015-09-22 10:12:03.820855

"""

# revision identifiers, used by Alembic.
revision = '2fb76f1a1393'
down_revision = '1ff737bef438'

from alembic import context
from alembic import op
import sqlalchemy as sa
from sqlalchemy.engine import reflection


def upgrade():
    migration_context = context.get_context()
    insp = reflection.Inspector(migration_context.bind)
    indx_names = [x['name'] for x in insp.get_indexes('test_metadata')]
    # Prempt duplicate index creation on sqlite
    if migration_context.dialect.name == 'sqlite':
        if 'ix_test_key_value' in indx_names:
            op.drop_index('ix_test_key_value', 'test_metadata')
    # NOTE(mtreinish) on some mysql versions renaming the column with a fk
    # constraint errors out so, delete it before the rename and add it back
    # after
    if migration_context.dialect.name == 'mysql':
        op.drop_constraint('test_metadata_ibfk_1', 'test_metadata',
                           'foreignkey')
    with op.batch_alter_table('test_metadata') as batch_op:
        batch_op.alter_column('test_run_id',
                              existing_type=sa.String(36),
                              existing_nullable=False,
                              new_column_name='test_id')
    if migration_context.dialect.name == 'mysql':
        op.create_foreign_key('test_metadata_ibfk_1', 'test_metadata',
                              'tests', ["test_id"], ['id'])


def downgrade():
    pass



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add microsecond columns to test_runs table

Revision ID: 1679b5bc102c
Revises: 5332fe255095
Create Date: 2015-02-27 18:39:13.275801

"""

# revision identifiers, used by Alembic.
revision = '1679b5bc102c'
down_revision = '5332fe255095'

import os

from alembic import context
from alembic import op
from oslo_config import cfg
import sqlalchemy as sa


CONF = cfg.CONF


def upgrade():
    migration_file = ('1679b5bc102c_add_subsecond_columns_to_test_runs_table.'
                      'mysql_upgrade.sql')
    migration_dir = os.path.dirname(os.path.realpath(__file__))
    sql_path = os.path.join(migration_dir, migration_file)
    migration_context = context.get_context()
    if migration_context.dialect.name == 'mysql':
        with open(sql_path, 'r') as sql_file:
            for line in sql_file.read().splitlines():
                # don't execute empty or commented lines
                if line and not line.startswith('--'):
                    op.execute(line)
    else:
        op.add_column('test_runs', sa.Column('start_time_microsecond',
                                             sa.Integer(), default=0))
        op.add_column('test_runs', sa.Column('stop_time_microsecond',
                                             sa.Integer(), default=0))
        if not CONF.disable_microsecond_data_migration:
            bind = op.get_bind()
            metadata = sa.schema.MetaData()
            metadata.bind = bind
            test_runs = sa.Table('test_runs', metadata, autoload=True)
            res = test_runs.select().execute()
            for test_run in res:
                start_micro = test_run[4].microsecond
                stop_micro = test_run[5].microsecond
                values = {'start_time_microsecond': start_micro,
                          'stop_time_microsecond': stop_micro}
                op.execute(test_runs.update().where(
                    test_runs.c.id == test_run[0]).values(values))
            res.close()


def downgrade():
    op.drop_column('test_runs', 'stop_time_microsecond')
    op.drop_column('test_runs', 'start_time_microsecond')



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""create test_runs table

Revision ID: 3db7b49816d5
Revises: 1f92cfe8a6d3
Create Date: 2014-06-08 14:34:56.786781

"""

# revision identifiers, used by Alembic.
revision = '3db7b49816d5'
down_revision = '1f92cfe8a6d3'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('test_runs',
                    sa.Column('id', sa.String(36), primary_key=True),
                    sa.Column('test_id', sa.String(36),
                              sa.ForeignKey('tests.id'),
                              nullable=False, index=True),
                    sa.Column('run_id', sa.String(36),
                              sa.ForeignKey('runs.id'),
                              nullable=False, index=True),
                    sa.Column('status', sa.String(256)),
                    sa.Column('start_time', sa.DateTime()),
                    sa.Column('stop_time', sa.DateTime()),
                    mysql_engine='InnoDB')


def downgrade():
    op.drop_table('test_runs')



# Copyright 2010-2011 OpenStack Foundation
# Copyright 2013 IBM Corp.
# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import os
import uuid

from alembic import config
from alembic import script
import six
from six.moves import configparser as ConfigParser
import sqlalchemy
from sqlalchemy.engine import reflection

from subunit2sql import exceptions as exc
from subunit2sql.tests import base
from subunit2sql.tests import db_test_utils
from subunit2sql.tests import subunit2sql_fixtures as fixtures


def get_table(engine, name):
    """Returns an sqlalchemy table dynamically from db.

    Needed because the models don't work for us in migrations
    as models will be far out of sync with the current data.
    """
    metadata = sqlalchemy.schema.MetaData()
    metadata.bind = engine
    return sqlalchemy.Table(name, metadata, autoload=True)


class TestWalkMigrations(base.TestCase):

    DEFAULT_CONFIG_FILE = os.path.join(os.path.dirname(__file__),
                                       'test_migrations.conf')
    CONFIG_FILE_PATH = os.environ.get('SUBUNIT2SQL_TEST_MIGRATIONS_CONF',
                                      DEFAULT_CONFIG_FILE)
    script_location = os.path.join(os.path.dirname(os.path.dirname(
        os.path.dirname(__file__))), 'migrations')

    def setUp(self):
        super(TestWalkMigrations, self).setUp()

        self.snake_walk = False
        self.test_databases = {}

        if os.path.exists(self.CONFIG_FILE_PATH):
            cp = ConfigParser.RawConfigParser()
            try:
                cp.read(self.CONFIG_FILE_PATH)
                defaults = cp.options('unit_tests')
                for key in defaults:
                    self.test_databases[key] = cp.get('unit_tests', key)
                self.snake_walk = cp.getboolean('walk_style', 'snake_walk')
            except ConfigParser.ParsingError as e:
                self.fail("Failed to read test_migrations.conf config "
                          "file. Got error: %s" % e)
        else:
            self.fail("Failed to find test_migrations.conf config "
                      "file.")

        self.engines = {}
        for key, value in self.test_databases.items():
            self.engines[key] = sqlalchemy.create_engine(value)

    def assertColumnExists(self, engine, table, column):
        table = get_table(engine, table)
        self.assertIn(column, table.c)

    def _revisions(self):
        """Provides revisions and its parent revisions.

        :return: List of tuples. Every tuple contains revision and its parent
        revision.
        """
        db_config = config.Config(os.path.join(self.script_location,
                                               'alembic.ini'))
        db_config.set_main_option('script_location', 'subunit2sql:migrations')
        script_dir = script.ScriptDirectory.from_config(db_config)
        revisions = list(script_dir.walk_revisions("base", "head"))

        if not revisions:
            raise exc.DbMigrationError('There is no suitable migrations.')

        for rev in list(reversed(revisions)):
            # Destination, current
            yield rev.revision, rev.down_revision

    def _walk_versions(self, engine):
        """Test migrations ability to upgrade."""

        revisions = self._revisions()
        for dest, curr in revisions:
            self._migrate_up(engine, dest, curr, with_data=True)

    def _migrate_up(self, engine, dest, curr, with_data=False):
        if with_data:
            data = None
            pre_upgrade = getattr(
                self, "_pre_upgrade_%s" % dest, None)
            if pre_upgrade:
                data = pre_upgrade(engine)
        db_test_utils.run_migration(dest, engine)
        if with_data:
            check = getattr(self, "_check_%s" % dest, None)
            if check and data:
                check(engine, data)

    def test_walk_versions(self):
        """Test walk versions.

        Walks all version scripts for each tested database, ensuring
        that there are no errors in the version scripts for each engine
        """
        for key, engine in self.engines.items():
            self._walk_versions(engine)

    def test_mysql_connect_fail(self):
        """Test graceful mysql connection failure.

        Test that we can trigger a mysql connection failure and we fail
        gracefully to ensure we don't break people without mysql
        """
        if db_test_utils.is_backend_avail('mysql', user="openstack_cifail"):
            self.fail("Shouldn't have connected")

    def test_mysql_opportunistically(self):
        self.useFixture(fixtures.LockFixture('mysql'))
        if not db_test_utils.is_backend_avail('mysql'):
            raise self.skipTest('mysql is not available')

        self.useFixture(fixtures.MySQLConfFixture())
        # Test that table creation on mysql only builds InnoDB tables
        # add this to the global lists to make reset work with it, it's removed
        # automatically in tearDown so no need to clean it up here.
        connect_string = db_test_utils.get_connect_string("mysql")
        engine = sqlalchemy.create_engine(connect_string)
        self.engines["mysqlcitest"] = engine
        self.test_databases["mysqlcitest"] = connect_string

        # build a fully populated mysql database with all the tables
        self._walk_versions(engine)

        connection = engine.connect()
        # sanity check
        total = connection.execute("SELECT count(*) "
                                   "from information_schema.TABLES "
                                   "where TABLE_SCHEMA='openstack_citest'")
        self.assertTrue(total.scalar() > 0, "No tables found. Wrong schema?")

        noninnodb = connection.execute("SELECT count(*) "
                                       "from information_schema.TABLES "
                                       "where TABLE_SCHEMA='openstack_citest' "
                                       "and ENGINE!='InnoDB' "
                                       "and TABLE_NAME!='alembic_version'")
        count = noninnodb.scalar()
        self.assertEqual(count, 0, "%d non InnoDB tables created" % count)
        connection.close()

    def test_postgresql_connect_fail(self):
        """Test graceful postgresql connection failure.

        Test that we can trigger a postgres connection failure and we fail
        gracefully to ensure we don't break people without postgres
        """
        if db_test_utils.is_backend_avail('postgresql',
                                          user="openstack_cifail"):
            self.fail("Shouldn't have connected")

    def test_postgresql_opportunistically(self):
        # Test postgresql database migration walk
        self.useFixture(fixtures.LockFixture('postgres'))
        if not db_test_utils.is_backend_avail('postgres'):
            raise self.skipTest('postgres is not available')
        self.useFixture(fixtures.PostgresConfFixture())
        # add this to the global lists to make reset work with it, it's removed
        # automatically in tearDown so no need to clean it up here.
        connect_string = db_test_utils.get_connect_string("postgres")
        engine = sqlalchemy.create_engine(connect_string)
        self.engines["postgresqlcitest"] = engine
        self.test_databases["postgresqlcitest"] = connect_string

        # build a fully populated postgresql database with all the tables
        self._walk_versions(engine)

    def test_sqlite_opportunistically(self):
        self.useFixture(fixtures.LockFixture('sqlite'))
        self.useFixture(fixtures.SqliteConfFixture())

        connect_string = db_test_utils.get_connect_string("sqlite")
        engine = sqlalchemy.create_engine(connect_string)
        self.engines["sqlitecitest"] = engine
        self.test_databases["sqlitecitest"] = connect_string

        self._walk_versions(engine)

    def _pre_upgrade_1f92cfe8a6d3(self, engine):
        tests = get_table(engine, 'tests')
        data = {'id': 'fake_test.id',
                'test_id': 'fake_project.tests.FakeTestClass.fake_test',
                'run_count': 1,
                'success': 1,
                'failure': 0}
        tests.insert().values(data).execute()
        return data

    def _pre_upgrade_3db7b49816d5(self, engine):
        runs = get_table(engine, 'runs')
        data = {'id': 'fake_run.id',
                'skips': 0,
                'fails': 0,
                'passes': 1,
                'run_time': 1.0,
                'artifacts': 'fake_url'}
        runs.insert().values(data).execute()
        return data

    def _pre_upgrade_163fd5aa1380(self, engine):
        now = datetime.datetime.now()
        test_runs = get_table(engine, 'test_runs')
        data = {'id': 'fake_test_run.id',
                'test_id': 'fake_test.id',
                'run_id': 'fake_run.id',
                'status': 'pass',
                'start_time': now,
                'stop_time': now + datetime.timedelta(1, 0)}
        test_runs.insert().values(data).execute()
        return data

    def _check_163fd5aa1380(self, engine, data):
        self.assertColumnExists(engine, 'tests', 'run_time')

    def _pre_upgrade_28ac1ba9c3db(self, engine):
        runs = get_table(engine, 'runs')
        data = [{'id': 'fake_run.id1',
                 'skips': 0,
                 'fails': 0,
                 'passes': 1,
                 'run_time': 1.0,
                 'artifacts': 'fake_url'},
                {'id': 'fake_run.id2',
                 'skips': 0,
                 'fails': 0,
                 'passes': 1,
                 'run_time': 1.0,
                 'artifacts': 'fake_url'}]
        for run in data:
            runs.insert().values(run).execute()
        return data

    def _check_28ac1ba9c3db(self, engine, data):
        self.assertColumnExists(engine, 'runs', 'run_at')
        runs = get_table(engine, 'runs')
        now = datetime.datetime.now().replace(microsecond=0)
        time_data = {'id': 'fake_run_with_time.id1',
                     'skips': 0,
                     'fails': 0,
                     'passes': 1,
                     'run_time': 1.0,
                     'artifacts': 'fake_url',
                     'run_at': now}
        runs.insert().values(time_data).execute()
        runs = get_table(engine, 'runs')
        result = runs.select().execute()
        run_at = list(map(lambda x: (x['id'], x['run_at']), result))
        for run in data:
            self.assertIn((run['id'], None), run_at)
        self.assertIn((time_data['id'], now), run_at)

    def _pre_upgrade_5332fe255095(self, engine):
        tests = get_table(engine, 'tests')
        test_runs = get_table(engine, 'test_runs')
        # Create 2 sample rows one for a passing test the other failing
        fake_tests = {'pass': {'id': 'fake_null_test_id',
                               'test_id': 'I_am_a_little_test_that_works',
                               'success': 2,
                               'failure': 0},
                      'fail': {'id': 'fake_null_test_id_fails',
                               'test_id': 'Im_a_little_test_that_doesnt_work',
                               'success': 0,
                               'failure': 1}}
        now = datetime.datetime.now()
        future_now = now + datetime.timedelta(0, 4)
        # Create sample rows for the test_runs corresponding to the test rows
        fake_test_runs = {'pass': [
            {'id': 'fake_test_run_pass_1', 'test_id': 'fake_null_test_id',
             'run_id': 'fake_run.id1', 'start_time': now, 'status': 'success',
             'stop_time': future_now},
            {'id': 'fake_test_run_pass_2', 'test_id': 'fake_null_test_id',
             'run_id': 'fake_run.id2', 'start_time': now, 'status': 'success',
             'stop_time': future_now}]}
        fake_test_runs['fail'] = {'id': 'fake_test_run_fail',
                                  'test_id': 'fake_null_test_id_fails',
                                  'run_id': 'fake_run.id1',
                                  'start_time': now,
                                  'status': 'fail',
                                  'stop_time': future_now}
        for test in fake_tests:
            tests.insert().values(fake_tests[test]).execute()
        for test_run in fake_test_runs['pass']:
            test_runs.insert().values(test_run).execute()
        test_runs.insert().values(fake_test_runs['fail']).execute()
        return {'tests': fake_tests, 'test_runs': fake_test_runs}

    def _check_5332fe255095(self, engine, data):
        tests = get_table(engine, 'tests')
        # Get the test uuids from the same data set
        test_ids = [data['tests'][x]['id'] for x in data['tests']]
        # Query the DB for the tests from the sample dataset above
        where = ' OR '.join(["tests.id='%s'" % x for x in test_ids])
        result = tests.select(sqlalchemy.sql.text(where)).execute()
        run_time_pairs = list(map(lambda x: (x['id'], x['run_time']), result))
        # Ensure the test with one failure is None
        self.assertIn(('fake_null_test_id_fails', None), run_time_pairs)
        # Ensure the test with 2 success each taking 4 sec lists the proper
        # run_time
        self.assertIn(('fake_null_test_id', 4.0), run_time_pairs)

    def _pre_upgrade_1679b5bc102c(self, engine):
        test_runs = get_table(engine, 'test_runs')
        now = datetime.datetime.now()
        future_now = now + datetime.timedelta(0, 4)
        fake_test_runs = {'id': 'abc123',
                          'test_id': 'fake_null_test_id',
                          'run_id': 'fake_run.id',
                          'status': 'success',
                          'start_time': now,
                          'stop_time': future_now}
        test_runs.insert().values(fake_test_runs).execute()
        return fake_test_runs

    def _check_1679b5bc102c(self, engine, data):
        test_runs = get_table(engine, 'test_runs')
        start_micro = data['start_time'].microsecond
        stop_micro = data['stop_time'].microsecond
        result = test_runs.select().execute()
        row = None
        for i in result:
            if i.id == data['id']:
                row = i
                break
        else:
            self.fail("Row not present")
        if row.start_time_microsecond == 0 and row.stop_time_microsecond == 0:
            # Backing db doesn't store subseconds so the migration will just
            # populate zeros and the data is lost to the ether.
            pass
        else:
            self.assertEqual(start_micro, row.start_time_microsecond)
            self.assertEqual(row.stop_time_microsecond, stop_micro)

    def _pre_upgrade_2fb76f1a1393(self, engine):
        test_metadata = get_table(engine, 'test_metadata')
        tests = get_table(engine, 'tests')
        test = {'id': 'fake_test_with_metadata.id',
                'test_id': 'fake_project.tests.FakeTestClass.fake_test_meta',
                'run_count': 1,
                'success': 1,
                'failure': 0}
        tests.insert().values(test).execute()
        data = {'id': 'AUUID',
                'key': 'AKey',
                'value': 42,
                'test_run_id': 'fake_test_with_metadata.id'}
        test_metadata.insert().values(data).execute()
        return data

    def _check_2fb76f1a1393(self, engine, data):
        test_metadata = get_table(engine, 'test_metadata')
        res = list(test_metadata.select().execute())[0]
        self.assertEqual(res.id, data['id'])
        self.assertEqual(res.test_id, data['test_run_id'])

    def _pre_upgrade_b96122f780(self, engine):
        # NOTE(mtreinish) Return fake data to ensure we run check, this
        # is needed because the framework normall assumes you're preseeded
        # data is in the correct state post migration. But this time all we
        # want to do is ensure a single index exists so that isn't needed
        return 'data'

    def _check_b96122f780(self, engine, data):
        insp = reflection.Inspector(engine)
        indxs = insp.get_indexes('test_runs')
        # Check that we don't duplicate indexes anymore
        tests = [indx for indx in indxs if ['test_id'] == indx['column_names']]
        runs = [indx for indx in indxs if indx['column_names'] == ['run_id']]
        self.assertEqual(1, len(tests))
        self.assertEqual(1, len(runs))

    def _pre_upgrade_2822a408bdd0(self, engine):
        data = {}

        # Add run
        runs = get_table(engine, 'runs')
        run = {'id': six.text_type(uuid.uuid4()),
               'skips': 0,
               'fails': 0,
               'passes': 1,
               'run_time': 1.0,
               'artifacts': 'https://am_i_really_a_fake_url',
               'run_at': datetime.datetime.utcnow()}
        runs.insert().values(run).execute()
        data['run'] = run
        # Add test_metadata
        run_metadatas = get_table(engine, 'run_metadata')
        run_metadata = {'id': six.text_type(uuid.uuid4()),
                        'run_id': run['id'],
                        'key': 'attrs',
                        'value': 'an_attr'}
        run_metadatas.insert().values(run_metadata).execute()
        data['run_metadata'] = run_metadata

        # Add test
        tests = get_table(engine, 'tests')
        test = {'id': six.text_type(uuid.uuid4()),
                'test_id': 'I_am_a_real_test!',
                'success': 1,
                'failure': 0}
        tests.insert().values(test).execute()
        data['test'] = test

        # Add test_metadata
        test_metadatas = get_table(engine, 'test_metadata')
        test_metadata = {'id': six.text_type(uuid.uuid4()),
                         'test_id': test['id'],
                         'key': 'a_real_key',
                         'value': 'an_attr'}
        test_metadatas.insert().values(test_metadata).execute()
        data['test_metadata'] = test_metadata

        # Add test run
        now = datetime.datetime.now().replace(microsecond=0)
        future_now = now + datetime.timedelta(0, 4)

        test_runs = get_table(engine, 'test_runs')

        test_run = {'id': six.text_type(uuid.uuid4()),
                    'test_id': test['id'],
                    'run_id': run['id'],
                    'start_time': now,
                    'status': 'success',
                    'stop_time': future_now}
        test_runs.insert().values(test_run).execute()
        data['test_run'] = test_run

        # Add test_run_metadata
        test_run_metadatas = get_table(engine, 'test_run_metadata')
        test_run_metadata = {'id': six.text_type(uuid.uuid4()),
                             'test_run_id': test_run['id'],
                             'key': 'attrs',
                             'value': 'an_attr'}
        test_run_metadatas.insert().values(test_run_metadata).execute()
        data['test_run_metadata'] = test_run_metadata

        attachments = get_table(engine, 'attachments')
        attachment = {'id': six.text_type(uuid.uuid4()),
                      'test_run_id': test_run['id'],
                      'label': 'an_attachment',
                      'attachment': b'something'}
        attachments.insert().values(attachment).execute()
        data['attachment'] = attachment
        return data

    def _check_2822a408bdd0(self, engine, data):
        # Check Primary Keys
        insp = reflection.Inspector(engine)
        runs_pk = insp.get_pk_constraint('runs')
        self.assertEqual(['id'], runs_pk['constrained_columns'])
        run_meta_pk = insp.get_pk_constraint('run_metadata')
        self.assertEqual(['id'], run_meta_pk['constrained_columns'])
        tests_pk = insp.get_pk_constraint('tests')
        self.assertEqual(['id'], tests_pk['constrained_columns'])
        test_meta_pk = insp.get_pk_constraint('test_metadata')
        self.assertEqual(['id'], test_meta_pk['constrained_columns'])
        test_runs_pk = insp.get_pk_constraint('runs')
        self.assertEqual(['id'], test_runs_pk['constrained_columns'])
        test_run_meta_pk = insp.get_pk_constraint('tests')
        self.assertEqual(['id'], test_run_meta_pk['constrained_columns'])
        attach_pk = insp.get_pk_constraint('attachments')
        self.assertEqual(['id'], attach_pk['constrained_columns'])

        if engine.dialect.name == 'sqlite':
            new_id_type = sqlalchemy.Integer
        else:
            new_id_type = sqlalchemy.BigInteger

        # Check id column type
        runs_col = [x for x in insp.get_columns(
            'runs') if x['name'] == 'id'][0]
        self.assertIsInstance(runs_col['type'], new_id_type)
        run_meta_col = [x for x in insp.get_columns(
            'run_metadata') if x['name'] == 'id'][0]
        self.assertIsInstance(run_meta_col['type'], new_id_type)
        tests_col = [x for x in insp.get_columns(
            'tests') if x['name'] == 'id'][0]
        self.assertIsInstance(tests_col['type'], new_id_type)
        test_meta_col = [x for x in insp.get_columns(
            'test_metadata') if x['name'] == 'id'][0]
        self.assertIsInstance(test_meta_col['type'], new_id_type)
        test_runs_col = [x for x in insp.get_columns(
            'test_runs') if x['name'] == 'id'][0]
        self.assertIsInstance(test_runs_col['type'], new_id_type)
        test_run_meta_col = [x for x in insp.get_columns(
            'test_run_metadata') if x['name'] == 'id'][0]
        self.assertIsInstance(test_run_meta_col['type'], new_id_type)
        attach_col = [x for x in insp.get_columns(
            'attachments') if x['name'] == 'id'][0]
        self.assertIsInstance(attach_col['type'], new_id_type)

        # Check all the new ids match
        runs_t = get_table(engine, 'runs')
        run_ids = [x[1] for x in runs_t.select().execute()]
        run_metadatas_t = get_table(engine, 'run_metadata')
        tests_t = get_table(engine, 'tests')
        test_metadatas_t = get_table(engine, 'test_metadata')
        test_runs_t = get_table(engine, 'test_runs')
        test_runs_raw = list(test_runs_t.select().execute())
        test_run_test_ids = [x[1] for x in test_runs_raw]
        test_run_metadatas_t = get_table(engine, 'test_run_metadata')
        attachments_t = get_table(engine, 'attachments')
        # Get test we inserted before migration
        test_row = list(tests_t.select().where(
            tests_t.c.test_id == data['test']['test_id']).execute())[0]
        self.assertIn(test_row[0], test_run_test_ids)
        # Check test run
        test_run_row = list(test_runs_t.select().where(
            test_runs_t.c.test_id == test_row[0]).execute())[0]
        self.assertEqual(test_run_row[3], data['test_run']['status'])
        self.assertEqual(test_run_row[4].replace(microsecond=0),
                         data['test_run']['start_time'].replace(microsecond=0))
        self.assertEqual(test_run_row[5].replace(microsecond=0),
                         data['test_run']['stop_time'].replace(microsecond=0))
        self.assertIn(test_run_row[2], run_ids)
        # Check run
        run_row = list(runs_t.select().where(
            runs_t.c.id == test_run_row[2]).execute())[0]
        self.assertEqual(data['run']['artifacts'], run_row[6])
        self.assertEqual(data['run']['id'], run_row[0])
        # Check run metadata
        run_metadata_row = list(run_metadatas_t.select().where(
            run_metadatas_t.c.run_id == run_row[1]).execute())[0]
        self.assertEqual(data['run_metadata']['key'], run_metadata_row[1])
        self.assertEqual(data['run_metadata']['value'], run_metadata_row[2])
        # Check test metadata
        test_metadata_row = list(test_metadatas_t.select().where(
            test_metadatas_t.c.test_id == test_row[0]).execute())[0]
        self.assertEqual(data['test_metadata']['key'], test_metadata_row[1])
        self.assertEqual(data['test_metadata']['value'], test_metadata_row[2])
        # Check test run metadata
        test_run_metadata_row = list(test_run_metadatas_t.select().where(
            test_run_metadatas_t.c.test_run_id == test_run_row[0]).execute())
        test_run_metadata_row = test_run_metadata_row[0]
        self.assertEqual(data['test_run_metadata']['key'],
                         test_run_metadata_row[1])
        self.assertEqual(data['test_run_metadata']['value'],
                         test_run_metadata_row[2])
        # Check attachment
        attachment_row = list(attachments_t.select().where(
            attachments_t.c.test_run_id == test_run_row[0]).execute())[0]
        self.assertEqual(data['attachment']['label'], attachment_row[2])
        self.assertEqual(data['attachment']['attachment'], attachment_row[3])



# Copyright 2012 New Dream Network, LLC (DreamHost)
# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

from alembic import command as alembic_command
from alembic import config as alembic_config
from alembic import script as alembic_script
from alembic import util as alembic_util
from oslo_config import cfg
from oslo_db import options

from subunit2sql.db import api as db_api

HEAD_FILENAME = 'HEAD'


def state_path_def(*args):
    """Return an uninterpolated path relative to $state_path."""
    return os.path.join('$state_path', *args)


MIGRATION_OPTS = [
    cfg.BoolOpt('disable-microsecond-data-migration', short='d', default=False,
                help="If set to true this option will skip the data migration"
                     " part of the microsecond migration. The schema changes "
                     "will still be run. If the database has already stripped "
                     "out the microseconds from the timestamps this will skip "
                     "converting the microsecond field from the timestamps "
                     "into a separate column"),
]

CONF = cfg.CONF
CONF.register_cli_opts(options.database_opts, group='database')
CONF.register_cli_opts(MIGRATION_OPTS)
CONF.import_opt('verbose', 'subunit2sql.db.api')


def do_alembic_command(config, cmd, *args, **kwargs):
    try:
        getattr(alembic_command, cmd)(config, *args, **kwargs)
    except alembic_util.CommandError as e:
        alembic_util.err(str(e))


def do_check_migration(config, cmd):
    do_alembic_command(config, 'branches')
    validate_head_file(config)


def do_upgrade_downgrade(config, cmd):
    if not CONF.command.revision and not CONF.command.delta:
        raise SystemExit('You must provide a revision or relative delta')

    revision = CONF.command.revision

    if CONF.command.delta:
        sign = '+' if CONF.command.name == 'upgrade' else '-'
        revision = sign + str(CONF.command.delta)
    else:
        revision = CONF.command.revision

    do_alembic_command(config, cmd, revision, sql=CONF.command.sql)


def do_stamp(config, cmd):
    do_alembic_command(config, cmd,
                       CONF.command.revision,
                       sql=CONF.command.sql)


def do_revision(config, cmd):
    do_alembic_command(config, cmd,
                       message=CONF.command.message,
                       autogenerate=CONF.command.autogenerate,
                       sql=CONF.command.sql)
    update_head_file(config)


def validate_head_file(config):
    script = alembic_script.ScriptDirectory.from_config(config)
    if len(script.get_heads()) > 1:
        alembic_util.err('Timeline branches unable to generate timeline')

    head_path = os.path.join(script.versions, HEAD_FILENAME)
    if (os.path.isfile(head_path) and
        open(head_path).read().strip() == script.get_current_head()):
        return
    else:
        alembic_util.err('HEAD file does not match migration timeline head')


def expire_old(config, cmd):
    expire_age = int(CONF.command.expire_age)
    if not CONF.command.no_runs:
        print('Expiring old runs.')
        db_api.delete_old_runs(expire_age)
    if not CONF.command.no_test_runs:
        print('Expiring old test_runs')
        db_api.delete_old_test_runs(expire_age)


def update_head_file(config):
    script = alembic_script.ScriptDirectory.from_config(config)
    if len(script.get_heads()) > 1:
        alembic_util.err('Timeline branches unable to generate timeline')

    head_path = os.path.join(script.versions, HEAD_FILENAME)
    with open(head_path, 'w+') as f:
        f.write(script.get_current_head())


def add_command_parsers(subparsers):
    for name in ['current', 'history', 'branches']:
        parser = subparsers.add_parser(name)
        parser.set_defaults(func=do_alembic_command)

    parser = subparsers.add_parser('check_migration')
    parser.set_defaults(func=do_check_migration)

    for name in ['upgrade', 'downgrade']:
        parser = subparsers.add_parser(name)
        parser.add_argument('--delta', type=int)
        parser.add_argument('--sql', action='store_true')
        parser.add_argument('revision', nargs='?')
        parser.add_argument('--mysql-engine',
                            default='',
                            help='Change MySQL storage engine of current '
                                 'existing tables')
        parser.set_defaults(func=do_upgrade_downgrade)

    parser = subparsers.add_parser('stamp')
    parser.add_argument('--sql', action='store_true')
    parser.add_argument('revision')
    parser.set_defaults(func=do_stamp)

    parser = subparsers.add_parser('revision')
    parser.add_argument('-m', '--message')
    parser.add_argument('--autogenerate', action='store_true')
    parser.add_argument('--sql', action='store_true')
    parser.set_defaults(func=do_revision)

    parser = subparsers.add_parser('expire',
                                   help="delete old rows from runs and "
                                        "test_runs tables")
    parser.add_argument('--no-runs', action='store_true',
                        help="Don't delete any rows in the runs table")
    parser.add_argument('--no-test-runs', action='store_true',
                        help="Don't delete any rows in the test_runs table")
    parser.add_argument('--expire-age', '-e', default=186,
                        help="Number of days into the past to use as the "
                             "expiration point")
    parser.set_defaults(func=expire_old)

command_opt = cfg.SubCommandOpt('command',
                                title='Command',
                                help='Available commands',
                                handler=add_command_parsers)

CONF.register_cli_opt(command_opt)


def main():
    config = alembic_config.Config(os.path.join(os.path.dirname(__file__),
                                                'alembic.ini'))
    config.set_main_option('script_location',
                           'subunit2sql:migrations')
    config.subunit2sql_config = CONF
    CONF()
    CONF.command.func(config, CONF.command.name)

if __name__ == "__main__":
    sys.exit(main())
